//-----------------------------------------------------------------------------------
// Description : get 8 "asynchronous" buses @80MHz of FE-data,
//   synchronize, word-align and reformat into 24 channels @ 40MHz
//   + testing features (patterns, spy)
`timescale 1ns / 1ps

module InputStage ( 
  clk, clk2x, Reset, AsyncFifoRst, PatternSel, LatTrim,  
  Ch1_RX_D,   Ch1_RX_ER,    Ch1_RX_DV,    Ch1_RX_Clk,  
  Ch2_RX_D,   Ch2_RX_ER,    Ch2_RX_DV,    Ch2_RX_Clk,
  Ch3_RX_D,   Ch3_RX_ER,    Ch3_RX_DV,    Ch3_RX_Clk,
  Ch4_RX_D,   Ch4_RX_ER,    Ch4_RX_DV,    Ch4_RX_Clk,
  Ch5_RX_D,   Ch5_RX_ER,    Ch5_RX_DV,    Ch5_RX_Clk,
  Ch6_RX_D,   Ch6_RX_ER,    Ch6_RX_DV,    Ch6_RX_Clk,    
  Ch7_RX_D,   Ch7_RX_ER,    Ch7_RX_DV,    Ch7_RX_Clk,
  Ch8_RX_D,   Ch8_RX_ER,    Ch8_RX_DV,    Ch8_RX_Clk,
  WrEn,   RdEn,  ProgInputDelay,   FifoEmpty, FifoFull,  DLL_lock,
  Fifo1Status, Fifo2Status, Fifo3Status, Fifo4Status, 
  Fifo5Status, Fifo6Status, Fifo7Status, Fifo8Status, LW,
  qie1data,    qie2data,   qie3data,   qie4data,   qie5data,   qie6data,
  qie7data,    qie8data,   qie9data,   qie10data,  qie11data,  qie12data,
  qie13data,   qie14data,  qie15data,  qie16data,  qie17data,  qie18data,
  qie19data,   qie20data,  qie21data,  qie22data,  qie23data,  qie24data,
  Pattern1, Pattern2, Pattern3, Pattern4,
  Pattern5, Pattern6, Pattern7, Pattern8,
  InputSpyData1,  InputSpyData2,  InputSpyData3,  InputSpyData4,
  InputSpyData5,  InputSpyData6,  InputSpyData7,  InputSpyData8,
  Fib1BZero,      Fib2BZero,      Fib3BZero,      Fib4BZero, 
  Fib5BZero,      Fib6BZero,      Fib7BZero,      Fib8BZero,
  resetTP1,       resetTP2,       resetTP3,       resetTP4,
  resetTP5,       resetTP6,       resetTP7,       resetTP8,                     
  ErrAcc1, ErrAcc2, ErrAcc3, ErrAcc4, ErrAcc5, ErrAcc6, ErrAcc7, ErrAcc8   );
  
  input clk, clk2x, Reset, AsyncFifoRst, PatternSel, DLL_lock, LatTrim;
  input [15:0]  Ch1_RX_D,  Ch2_RX_D,  Ch3_RX_D,  Ch4_RX_D,  
                Ch5_RX_D,  Ch6_RX_D,  Ch7_RX_D,  Ch8_RX_D; 
  input    Ch1_RX_ER, Ch1_RX_DV, Ch1_RX_Clk,
           Ch2_RX_ER, Ch2_RX_DV, Ch2_RX_Clk,
           Ch3_RX_ER, Ch3_RX_DV, Ch3_RX_Clk,
           Ch4_RX_ER, Ch4_RX_DV, Ch4_RX_Clk,
           Ch5_RX_ER, Ch5_RX_DV, Ch5_RX_Clk,
           Ch6_RX_ER, Ch6_RX_DV, Ch6_RX_Clk,
           Ch7_RX_ER, Ch7_RX_DV, Ch7_RX_Clk,
           Ch8_RX_ER, Ch8_RX_DV, Ch8_RX_Clk;
  input  [17:0] Pattern1, Pattern2, Pattern3, Pattern4,
                Pattern5, Pattern6, Pattern7, Pattern8;
  input  [15:0]  ProgInputDelay; 
        
  output [10:0]
     qie1data,  qie2data,  qie3data,  qie4data,  qie5data,  qie6data, 
     qie7data,  qie8data,  qie9data,  qie10data, qie11data, qie12data,
     qie13data, qie14data, qie15data, qie16data, qie17data, qie18data,
     qie19data, qie20data, qie21data, qie22data, qie23data, qie24data;  
  output [17:0]  InputSpyData1,InputSpyData2,InputSpyData3,InputSpyData4, 
                 InputSpyData5,InputSpyData6,InputSpyData7,InputSpyData8;
  output [15:0]  ErrAcc1, ErrAcc2, ErrAcc3, ErrAcc4,
                 ErrAcc5, ErrAcc6, ErrAcc7, ErrAcc8;
  output  Fib1BZero, Fib2BZero, Fib3BZero, Fib4BZero, 
          Fib5BZero, Fib6BZero, Fib7BZero, Fib8BZero;  // one BZero per fiber
  output  resetTP1, resetTP2, resetTP3, resetTP4,
          resetTP5, resetTP6, resetTP7, resetTP8;
  output [8:1] FifoEmpty, FifoFull, WrEn, RdEn, LW;
  output [3:0] Fifo1Status, Fifo2Status, Fifo3Status, Fifo4Status, 
               Fifo5Status, Fifo6Status, Fifo7Status, Fifo8Status;   

  reg [15:0] ErrAcc1, ErrAcc2, ErrAcc3, ErrAcc4,
             ErrAcc5, ErrAcc6, ErrAcc7, ErrAcc8;     
  reg [10:0]
    qie1data,  qie2data,  qie3data,  qie4data,  qie5data,  qie6data, 
    qie7data,  qie8data,  qie9data,  qie10data, qie11data, qie12data,
    qie13data, qie14data, qie15data, qie16data, qie17data, qie18data,
    qie19data, qie20data, qie21data, qie22data, qie23data, qie24data; 
  reg  Fib1BZero, Fib2BZero, Fib3BZero, Fib4BZero, 
       Fib5BZero, Fib6BZero, Fib7BZero, Fib8BZero;  //it's one BZero per fiber
  reg [17:0]  InputSpyData1,InputSpyData2,InputSpyData3,InputSpyData4, 
              InputSpyData5,InputSpyData6,InputSpyData7,InputSpyData8;
  reg [15:0] ErrAccInt1, ErrAccInt2, ErrAccInt3, ErrAccInt4,
             ErrAccInt5, ErrAccInt6, ErrAccInt7, ErrAccInt8;
  reg [8:1]  FifoEmpty, empty_D, FifoFull, full_D;
  reg DL1, DL2, DL3, FifoRestart;
 
  wire [15:0]  Ch1_RX_Dw,  Ch2_RX_Dw,  Ch3_RX_Dw,  Ch4_RX_Dw,  
               Ch5_RX_Dw,  Ch6_RX_Dw,  Ch7_RX_Dw,  Ch8_RX_Dw; 
  wire     Ch1_RX_ERw, Ch1_RX_DVw,  Ch2_RX_ERw, Ch2_RX_DVw,
           Ch3_RX_ERw, Ch3_RX_DVw,  Ch4_RX_ERw, Ch4_RX_DVw,
           Ch5_RX_ERw, Ch5_RX_DVw,  Ch6_RX_ERw, Ch6_RX_DVw,
           Ch7_RX_ERw, Ch7_RX_DVw,  Ch8_RX_ERw, Ch8_RX_DVw; 
  wire [17:0]
   Ch1_Input_Dataw,Ch2_Input_Dataw,Ch3_Input_Dataw,Ch4_Input_Dataw,
   Ch5_Input_Dataw,Ch6_Input_Dataw,Ch7_Input_Dataw,Ch8_Input_Dataw,
   Ch1Aligned, Ch2Aligned, Ch3Aligned, Ch4Aligned,
   Ch5Aligned, Ch6Aligned, Ch7Aligned, Ch8Aligned;
  wire [10:0]
    qie1dataw,  qie2dataw,  qie3dataw,  qie4dataw,  qie5dataw,  qie6dataw, 
    qie7dataw,  qie8dataw,  qie9dataw,  qie10dataw, qie11dataw, qie12dataw,
    qie13dataw, qie14dataw, qie15dataw, qie16dataw, qie17dataw, qie18dataw,
    qie19dataw, qie20dataw, qie21dataw, qie22dataw, qie23dataw, qie24dataw;
  wire  Fib1BZeroW, Fib2BZeroW, Fib3BZeroW, Fib4BZeroW, 
        Fib5BZeroW, Fib6BZeroW, Fib7BZeroW, Fib8BZeroW;
  wire [8:1] empty, full;

  always @ (posedge clk2x) begin 
 // We generate a 3-tick long FifoRestart, so it will "cover" 
 //   several  states of the FifoState FSM.
      DL1 <= DLL_lock;  DL2 <= DL1;  DL3 <= DL2;
      FifoRestart <= DLL_lock & (~DL3); 
  end
  
 FifoStage FifoStage 
 (.clk2x(clk2x), .AsyncFifoRst(AsyncFifoRst), .FifoRestart(FifoRestart),
  .LatTrim(LatTrim), .WrEn(WrEn), .RdEn(RdEn), .full(full), .LW(LW),
  .Fifo1Status(Fifo1Status),    .Fifo2Status(Fifo2Status),  .empty(empty), 
  .Fifo3Status(Fifo3Status),    .Fifo4Status(Fifo4Status), 
  .Fifo5Status(Fifo5Status),    .Fifo6Status(Fifo6Status), 
  .Fifo7Status(Fifo7Status),    .Fifo8Status(Fifo8Status), 
  .Ch1RxD(Ch1_RX_D),  .Ch1RxEr(Ch1_RX_ER),   .Ch1RxDV(Ch1_RX_DV),  .Ch1_RX_Clk(Ch1_RX_Clk),  
  .Ch2RxD(Ch2_RX_D),  .Ch2RxEr(Ch2_RX_ER),   .Ch2RxDV(Ch2_RX_DV),  .Ch2_RX_Clk(Ch2_RX_Clk),
  .Ch3RxD(Ch3_RX_D),  .Ch3RxEr(Ch3_RX_ER),   .Ch3RxDV(Ch3_RX_DV),  .Ch3_RX_Clk(Ch3_RX_Clk),
  .Ch4RxD(Ch4_RX_D),  .Ch4RxEr(Ch4_RX_ER),   .Ch4RxDV(Ch4_RX_DV),  .Ch4_RX_Clk(Ch4_RX_Clk),
  .Ch5RxD(Ch5_RX_D),  .Ch5RxEr(Ch5_RX_ER),   .Ch5RxDV(Ch5_RX_DV),  .Ch5_RX_Clk(Ch5_RX_Clk),
  .Ch6RxD(Ch6_RX_D),  .Ch6RxEr(Ch6_RX_ER),   .Ch6RxDV(Ch6_RX_DV),  .Ch6_RX_Clk(Ch6_RX_Clk),    
  .Ch7RxD(Ch7_RX_D),  .Ch7RxEr(Ch7_RX_ER),   .Ch7RxDV(Ch7_RX_DV),  .Ch7_RX_Clk(Ch7_RX_Clk),
  .Ch8RxD(Ch8_RX_D),  .Ch8RxEr(Ch8_RX_ER),   .Ch8RxDV(Ch8_RX_DV),  .Ch8_RX_Clk(Ch8_RX_Clk),
  .Ch1RxD_o(Ch1_RX_Dw),   .Ch1RxEr_o(Ch1_RX_ERw),    .Ch1RxDV_o(Ch1_RX_DVw),      
  .Ch2RxD_o(Ch2_RX_Dw),   .Ch2RxEr_o(Ch2_RX_ERw),    .Ch2RxDV_o(Ch2_RX_DVw),    
  .Ch3RxD_o(Ch3_RX_Dw),   .Ch3RxEr_o(Ch3_RX_ERw),    .Ch3RxDV_o(Ch3_RX_DVw),   
  .Ch4RxD_o(Ch4_RX_Dw),   .Ch4RxEr_o(Ch4_RX_ERw),    .Ch4RxDV_o(Ch4_RX_DVw),    
  .Ch5RxD_o(Ch5_RX_Dw),   .Ch5RxEr_o(Ch5_RX_ERw),    .Ch5RxDV_o(Ch5_RX_DVw),    
  .Ch6RxD_o(Ch6_RX_Dw),   .Ch6RxEr_o(Ch6_RX_ERw),    .Ch6RxDV_o(Ch6_RX_DVw),        
  .Ch7RxD_o(Ch7_RX_Dw),   .Ch7RxEr_o(Ch7_RX_ERw),    .Ch7RxDV_o(Ch7_RX_DVw),    
  .Ch8RxD_o(Ch8_RX_Dw),   .Ch8RxEr_o(Ch8_RX_ERw),    .Ch8RxDV_o(Ch8_RX_DVw)    ); 
  
 DelayStage8Ch DelayStage8Ch (
  .Ch1DIN({Ch1_RX_ERw, Ch1_RX_DVw, Ch1_RX_Dw}),
  .Ch2DIN({Ch2_RX_ERw, Ch2_RX_DVw, Ch2_RX_Dw}), 
  .Ch3DIN({Ch3_RX_ERw, Ch3_RX_DVw, Ch3_RX_Dw}), 
  .Ch4DIN({Ch4_RX_ERw, Ch4_RX_DVw, Ch4_RX_Dw}), 
  .Ch5DIN({Ch5_RX_ERw, Ch5_RX_DVw, Ch5_RX_Dw}), 
  .Ch6DIN({Ch6_RX_ERw, Ch6_RX_DVw, Ch6_RX_Dw}), 
  .Ch7DIN({Ch7_RX_ERw, Ch7_RX_DVw, Ch7_RX_Dw}), 
  .Ch8DIN({Ch8_RX_ERw, Ch8_RX_DVw, Ch8_RX_Dw}), 
  .Ch1Delay(ProgInputDelay[1:0]), .Ch2Delay(ProgInputDelay[3:2]), 
  .Ch3Delay(ProgInputDelay[5:4]), .Ch4Delay(ProgInputDelay[7:6]), 
  .Ch5Delay(ProgInputDelay[9:8]), .Ch6Delay(ProgInputDelay[11:10]), 
  .Ch7Delay(ProgInputDelay[13:12]), .Ch8Delay(ProgInputDelay[15:14]), 
  .Ch1DOUT(Ch1Aligned), .Ch2DOUT(Ch2Aligned), .Ch3DOUT(Ch3Aligned), .Ch4DOUT(Ch4Aligned), 
  .Ch5DOUT(Ch5Aligned), .Ch6DOUT(Ch6Aligned), .Ch7DOUT(Ch7Aligned), .Ch8DOUT(Ch8Aligned),
  .CLK(clk2x)             ); 
  
  data_select_stage  mux_inst
  (  .SysClkx2(clk2x),  .sel(PatternSel), 
  .Ch1_RX_D(Ch1Aligned[15:0]),  .Ch1_RX_ER(Ch1Aligned[17]),  .Ch1_RX_DV(Ch1Aligned[16]),      
  .Ch2_RX_D(Ch2Aligned[15:0]),  .Ch2_RX_ER(Ch2Aligned[17]),  .Ch2_RX_DV(Ch2Aligned[16]),    
  .Ch3_RX_D(Ch3Aligned[15:0]),  .Ch3_RX_ER(Ch3Aligned[17]),  .Ch3_RX_DV(Ch3Aligned[16]),    
  .Ch4_RX_D(Ch4Aligned[15:0]),  .Ch4_RX_ER(Ch4Aligned[17]),  .Ch4_RX_DV(Ch4Aligned[16]),    
  .Ch5_RX_D(Ch5Aligned[15:0]),  .Ch5_RX_ER(Ch5Aligned[17]),  .Ch5_RX_DV(Ch5Aligned[16]),    
  .Ch6_RX_D(Ch6Aligned[15:0]),  .Ch6_RX_ER(Ch6Aligned[17]),  .Ch6_RX_DV(Ch6Aligned[16]),        
  .Ch7_RX_D(Ch7Aligned[15:0]),  .Ch7_RX_ER(Ch7Aligned[17]),  .Ch7_RX_DV(Ch7Aligned[16]),    
  .Ch8_RX_D(Ch8Aligned[15:0]),  .Ch8_RX_ER(Ch8Aligned[17]),  .Ch8_RX_DV(Ch8Aligned[16]),    
  .Pattern1(Pattern1),  .Pattern2(Pattern2),  .Pattern3(Pattern3),  .Pattern4(Pattern4), 
  .Pattern5(Pattern5),  .Pattern6(Pattern6),  .Pattern7(Pattern7),  .Pattern8(Pattern8), 
  .Ch1_Input_Data_o(Ch1_Input_Dataw),    .Ch2_Input_Data_o(Ch2_Input_Dataw),    
  .Ch3_Input_Data_o(Ch3_Input_Dataw),    .Ch4_Input_Data_o(Ch4_Input_Dataw),     
  .Ch5_Input_Data_o(Ch5_Input_Dataw),    .Ch6_Input_Data_o(Ch6_Input_Dataw),          
  .Ch7_Input_Data_o(Ch7_Input_Dataw),    .Ch8_Input_Data_o(Ch8_Input_Dataw)         );
  
  clk_change_stage   clk_change_stage 
  ( .clk2x(clk2x),              .clk(clk),
   .Input_Data1(Ch1_Input_Dataw),  .Input_Data2(Ch2_Input_Dataw),
   .Input_Data3(Ch3_Input_Dataw),  .Input_Data4(Ch4_Input_Dataw),
   .Input_Data5(Ch5_Input_Dataw),  .Input_Data6(Ch6_Input_Dataw), 
   .Input_Data7(Ch7_Input_Dataw),  .Input_Data8(Ch8_Input_Dataw),
   .QIE1data(qie1dataw),   .QIE2data(qie2dataw),    .QIE3data(qie3dataw),
   .QIE4data(qie4dataw),   .QIE5data(qie5dataw),    .QIE6data(qie6dataw),
   .QIE7data(qie7dataw),   .QIE8data(qie8dataw),    .QIE9data(qie9dataw),
   .QIE10data(qie10dataw), .QIE11data(qie11dataw),  .QIE12data(qie12dataw),
   .QIE13data(qie13dataw), .QIE14data(qie14dataw),  .QIE15data(qie15dataw),
   .QIE16data(qie16dataw), .QIE17data(qie17dataw),  .QIE18data(qie18dataw),
   .QIE19data(qie19dataw), .QIE20data(qie20dataw),  .QIE21data(qie21dataw),
   .QIE22data(qie22dataw), .QIE23data(qie23dataw),  .QIE24data(qie24dataw),
   .Fib1BZero(Fib1BZeroW), .Fib2BZero(Fib2BZeroW),  .Fib3BZero(Fib3BZeroW),  .Fib4BZero(Fib4BZeroW), 
   .Fib5BZero(Fib5BZeroW), .Fib6BZero(Fib6BZeroW),  .Fib7BZero(Fib7BZeroW),  .Fib8BZero(Fib8BZeroW),
   .resetTP1(resetTP1), .resetTP2(resetTP2), .resetTP3(resetTP3), .resetTP4(resetTP4),
   .resetTP5(resetTP5), .resetTP6(resetTP6), .resetTP7(resetTP7), .resetTP8(resetTP8)         );
    
  always@(posedge clk)
   begin 
 // output registers as signals go to 2 destinations, one is a RAM address (slow ?)
    qie1data  <= qie1dataw;          qie2data  <= qie2dataw;
    qie3data  <= qie3dataw;          qie4data  <= qie4dataw;
    qie5data  <= qie5dataw;          qie6data  <= qie6dataw;
    qie7data  <= qie7dataw;          qie8data  <= qie8dataw;
    qie9data  <= qie9dataw;          qie10data <= qie10dataw;
    qie11data <= qie11dataw;         qie12data <= qie12dataw;
    qie13data <= qie13dataw;         qie14data <= qie14dataw;
    qie15data <= qie15dataw;         qie16data <= qie16dataw;
    qie17data <= qie17dataw;         qie18data <= qie18dataw; 
    qie19data <= qie19dataw;         qie20data <= qie20dataw;
    qie21data <= qie21dataw;         qie22data <= qie22dataw;
    qie23data <= qie23dataw;         qie24data <= qie24dataw;

    Fib1BZero <= Fib1BZeroW;         Fib2BZero <= Fib2BZeroW;
    Fib3BZero <= Fib3BZeroW;         Fib4BZero <= Fib4BZeroW;
    Fib5BZero <= Fib5BZeroW;         Fib6BZero <= Fib6BZeroW;
    Fib7BZero <= Fib7BZeroW;         Fib8BZero <= Fib8BZeroW;
  end

  always@(posedge clk2x or posedge Reset) begin  // always
     if (Reset) begin  // reset  
        ErrAccInt1 <= 0;   ErrAccInt2 <= 0; 
        ErrAccInt3 <= 0;   ErrAccInt4 <= 0; 
        ErrAccInt5 <= 0;   ErrAccInt6 <= 0; 
        ErrAccInt7 <= 0;   ErrAccInt8 <= 0; 
        end   // reset
     else begin  // clk2x
       if (Ch1_RX_ERw & (ErrAccInt1 != 16'hFFFF))  ErrAccInt1 <= ErrAccInt1 + 1'b1;
       if (Ch2_RX_ERw & (ErrAccInt2 != 16'hFFFF))  ErrAccInt2 <= ErrAccInt2 + 1'b1;
       if (Ch3_RX_ERw & (ErrAccInt3 != 16'hFFFF))  ErrAccInt3 <= ErrAccInt3 + 1'b1;
       if (Ch4_RX_ERw & (ErrAccInt4 != 16'hFFFF))  ErrAccInt4 <= ErrAccInt4 + 1'b1;
       if (Ch5_RX_ERw & (ErrAccInt5 != 16'hFFFF))  ErrAccInt5 <= ErrAccInt5 + 1'b1;
       if (Ch6_RX_ERw & (ErrAccInt6 != 16'hFFFF))  ErrAccInt6 <= ErrAccInt6 + 1'b1;
       if (Ch7_RX_ERw & (ErrAccInt7 != 16'hFFFF))  ErrAccInt7 <= ErrAccInt7 + 1'b1;
       if (Ch8_RX_ERw & (ErrAccInt8 != 16'hFFFF))  ErrAccInt8 <= ErrAccInt8 + 1'b1;
     end  // clk2x
  end   //  always

  always@(posedge clk2x) begin
    InputSpyData1 <= Ch1_Input_Dataw; 
    InputSpyData2 <= Ch2_Input_Dataw;
    InputSpyData3 <= Ch3_Input_Dataw; 
    InputSpyData4 <= Ch4_Input_Dataw; 
    InputSpyData5 <= Ch5_Input_Dataw;
    InputSpyData6 <= Ch6_Input_Dataw;
    InputSpyData7 <= Ch7_Input_Dataw;
    InputSpyData8 <= Ch8_Input_Dataw;
    empty_D <= empty;     full_D  <= full; 
   end
               
  always@(posedge clk) begin	      
    // output registers at 40 MHz to improve Fmax
    ErrAcc1 <= ErrAccInt1;       ErrAcc2 <= ErrAccInt2;
    ErrAcc3 <= ErrAccInt3;       ErrAcc4 <= ErrAccInt4;
    ErrAcc5 <= ErrAccInt5;       ErrAcc6 <= ErrAccInt6;
    ErrAcc7 <= ErrAccInt7;       ErrAcc8 <= ErrAccInt8;
    // create 25-ns pulses from 12.5-ns pulses.
    // this also set to one a flag that keeps toggling. 
    FifoEmpty <= empty | empty_D;
    FifoFull  <= full  | full_D;
   end
                                                
endmodule
