/************************************************************************************/ 
// DESCRIPTION :   This module contains the L1A-LATENCY PIPELINES and derandomizer 
//  buffers for all Daq and TP channels. The pipeline RAM is logically a 
//  circular buffer used to delay the data waiting the L1A. 
//  The drnd buffers allow to average over time the rate of selected data.
//  NEW: different delay of Daq-data and TP-Data to offset delay of TriggerPath
`include "define.v"
module PipeDrnd
  (data_in_1_ch,   data_in_2_ch,    data_in_3_ch,    data_in_4_ch,
   data_in_5_ch,   data_in_6_ch,    data_in_7_ch,    data_in_8_ch,
   data_in_9_ch,   data_in_10_ch,   data_in_11_ch,   data_in_12_ch,
   data_in_13_ch,  data_in_14_ch,   data_in_15_ch,   data_in_16_ch,
   data_in_17_ch,  data_in_18_ch,   data_in_19_ch,   data_in_20_ch,
   data_in_21_ch,  data_in_22_ch,   data_in_23_ch,   data_in_24_ch,
   TP1_Data,   TP2_Data,  TP3_Data,   TP4_Data,  TP5_Data,   TP6_Data,
   TP7_Data,   TP8_Data,  TP9_Data,   TP10_Data, TP11_Data,  TP12_Data,
   TP13_Data,  TP14_Data, TP15_Data,  TP16_Data, TP17_Data,  TP18_Data, 
   TP19_Data,  TP20_Data, TP21_Data,  TP22_Data, TP23_Data,  TP24_Data,
   
   PipeLength,  DrndWrAddr,  DrndRdAddr,  DrndWe,   
   ChanSel,   DaqMuxOut,    NPS, clk, DInError, DInErDel    );

parameter
INPUT_DATA_WIDTH      = 11,
CIRC_BUF_AD_WIDTH     =  9;  

 input  clk, DInError; // DInError must experience the same delay of the data   
 input  [INPUT_DATA_WIDTH-1:0]   
             data_in_1_ch,   data_in_2_ch,   data_in_3_ch,   data_in_4_ch,
             data_in_5_ch,   data_in_6_ch,  data_in_7_ch,   data_in_8_ch,  
             data_in_9_ch,   data_in_10_ch,  data_in_11_ch,  data_in_12_ch,
             data_in_13_ch,  data_in_14_ch,  data_in_15_ch,  data_in_16_ch,
             data_in_17_ch,  data_in_18_ch,  data_in_19_ch,  data_in_20_ch,
             data_in_21_ch,  data_in_22_ch,  data_in_23_ch,  data_in_24_ch;
 input  [INPUT_DATA_WIDTH-1:0]
             TP1_Data,  TP2_Data,   TP3_Data,    TP4_Data,
             TP5_Data,  TP6_Data,   TP7_Data,    TP8_Data, 
             TP9_Data,  TP10_Data,  TP11_Data,   TP12_Data,
             TP13_Data, TP14_Data,  TP15_Data,   TP16_Data,
             TP17_Data, TP18_Data,  TP19_Data,   TP20_Data,
             TP21_Data, TP22_Data,  TP23_Data,   TP24_Data;
   
 input  [7:0] PipeLength;
 input  [4:0] NPS;
 input                          DrndWe;   
 input  [`DRND_ADDR_WIDTH-1:0]  DrndWrAddr,  DrndRdAddr;
 input  [`DAQ_MUX_SEL-1:0]      ChanSel;  
 output DInErDel; reg DInErDel;
 output [`MAIN_DATA_WIDTH -1:0]  DaqMuxOut;  reg [`MAIN_DATA_WIDTH -1:0] DaqMuxOut;  

 
  wire [35:0]   //connections from the L1A-LATENCY PIPELINES to the derandomizers 
   DaqCirBuf1Out,  DaqCirBuf2Out,  DaqCirBuf3Out,  DaqCirBuf4Out,
   DaqCirBuf5Out,  DaqCirBuf6Out,  DaqCirBuf7Out,  DaqCirBuf8Out,   
   TPCirBuf1Out,   TPCirBuf2Out,   TPCirBuf3Out,   TPCirBuf4Out,
   TPCirBuf5Out,   TPCirBuf6Out,   TPCirBuf7Out,   TPCirBuf8Out;

  wire  [35:0]   //connections from the derandomizer to the mux 
     DaqDrnd1Out,  DaqDrnd2Out,  DaqDrnd3Out,  DaqDrnd4Out,
     DaqDrnd5Out,  DaqDrnd6Out,  DaqDrnd7Out,  DaqDrnd8Out,  
     TpDrnd1Out,   TpDrnd2Out,   TpDrnd3Out,   TpDrnd4Out,
     TpDrnd5Out,   TpDrnd6Out,   TpDrnd7Out,   TpDrnd8Out ;

  reg  [CIRC_BUF_AD_WIDTH-1:0] CircBufWrAd, CircBufRdAd ;   
  wire [CIRC_BUF_AD_WIDTH-1:0] TPG_LATENCY = 'd3; //should match latency of TriggerPath 
    
   // L1A-LATENCY PIPELINES (= Circular Buffers)

  DP_Syn_RAM36x512    DaqCircularBuffer1
    (  clk,1'b1, CircBufWrAd + TPG_LATENCY, CircBufRdAd,
      {3'd0, data_in_3_ch,data_in_2_ch, data_in_1_ch},  DaqCirBuf1Out );
  DP_Syn_RAM36x512    DaqCircularBuffer2
    (  clk,1'b1, CircBufWrAd + TPG_LATENCY, CircBufRdAd,
      {3'd0, data_in_6_ch, data_in_5_ch, data_in_4_ch},  DaqCirBuf2Out );
  DP_Syn_RAM36x512    DaqCircularBuffer3
    (  clk,1'b1, CircBufWrAd + TPG_LATENCY, CircBufRdAd,
      {3'd0, data_in_9_ch, data_in_8_ch, data_in_7_ch},  DaqCirBuf3Out );
  DP_Syn_RAM36x512    DaqCircularBuffer4
    (  clk,1'b1, CircBufWrAd + TPG_LATENCY, CircBufRdAd,
      {3'd0, data_in_12_ch, data_in_11_ch, data_in_10_ch}, DaqCirBuf4Out );
  DP_Syn_RAM36x512    DaqCircularBuffer5
    (  clk,1'b1, CircBufWrAd + TPG_LATENCY, CircBufRdAd,
      {3'd0, data_in_15_ch, data_in_14_ch, data_in_13_ch}, DaqCirBuf5Out );
  DP_Syn_RAM36x512    DaqCircularBuffer6
    (  clk,1'b1, CircBufWrAd + TPG_LATENCY, CircBufRdAd,
      {3'd0, data_in_18_ch, data_in_17_ch, data_in_16_ch}, DaqCirBuf6Out );
  DP_Syn_RAM36x512    DaqCircularBuffer7
    (  clk,1'b1, CircBufWrAd + TPG_LATENCY, CircBufRdAd,
      {3'd0, data_in_21_ch, data_in_20_ch, data_in_19_ch}, DaqCirBuf7Out );
  DP_Syn_RAM36x512    DaqCircularBuffer8
    (  clk,1'b1, CircBufWrAd + TPG_LATENCY, CircBufRdAd,
      {3'd0, data_in_24_ch, data_in_23_ch, data_in_22_ch}, DaqCirBuf8Out );
   
  DP_Syn_RAM36x512    TpCircularBuffer1
    (  clk, 1'b1, CircBufWrAd, CircBufRdAd,
      {2'd0, DInError, TP3_Data, TP2_Data, TP1_Data},   TPCirBuf1Out );
  DP_Syn_RAM36x512    TpCircularBuffer2
    (  clk, 1'b1, CircBufWrAd, CircBufRdAd,
      {2'd0, DInError, TP6_Data, TP5_Data, TP4_Data},   TPCirBuf2Out );
  DP_Syn_RAM36x512    TpCircularBuffer3
    (  clk, 1'b1, CircBufWrAd, CircBufRdAd,
      {2'd0, DInError, TP9_Data, TP8_Data, TP7_Data},   TPCirBuf3Out );
  DP_Syn_RAM36x512    TpCircularBuffer4
    (  clk, 1'b1, CircBufWrAd, CircBufRdAd,
      {2'd0, DInError, TP12_Data, TP11_Data, TP10_Data}, TPCirBuf4Out );
  DP_Syn_RAM36x512    TpCircularBuffer5
    (  clk, 1'b1, CircBufWrAd, CircBufRdAd,
      {2'd0, DInError, TP15_Data, TP14_Data, TP13_Data}, TPCirBuf5Out );
  DP_Syn_RAM36x512    TpCircularBuffer6
    (  clk, 1'b1, CircBufWrAd, CircBufRdAd,
      {2'd0, DInError, TP18_Data, TP17_Data, TP16_Data}, TPCirBuf6Out );
  DP_Syn_RAM36x512    TpCircularBuffer7
    (  clk, 1'b1, CircBufWrAd, CircBufRdAd,
      {2'd0, DInError, TP21_Data, TP20_Data, TP19_Data}, TPCirBuf7Out );
  DP_Syn_RAM36x512    TpCircularBuffer8
    (  clk, 1'b1, CircBufWrAd, CircBufRdAd,
      {2'd0, DInError, TP24_Data, TP23_Data, TP22_Data}, TPCirBuf8Out );
   
// DERANDOMIZATION BUFFERS
    
  DP_Syn_RAM36x512    DaqDrnd1
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, DaqCirBuf1Out, DaqDrnd1Out);   
  DP_Syn_RAM36x512    DaqDrnd2
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, DaqCirBuf2Out, DaqDrnd2Out);
  DP_Syn_RAM36x512    DaqDrnd3
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, DaqCirBuf3Out, DaqDrnd3Out);
  DP_Syn_RAM36x512    DaqDrnd4
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, DaqCirBuf4Out, DaqDrnd4Out);
  DP_Syn_RAM36x512    DaqDrnd5
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, DaqCirBuf5Out, DaqDrnd5Out);
  DP_Syn_RAM36x512    DaqDrnd6
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, DaqCirBuf6Out, DaqDrnd6Out);
  DP_Syn_RAM36x512    DaqDrnd7
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, DaqCirBuf7Out, DaqDrnd7Out);
  DP_Syn_RAM36x512    DaqDrnd8
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, DaqCirBuf8Out, DaqDrnd8Out);

  DP_Syn_RAM36x512    TpDrnd1
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, TPCirBuf1Out, TpDrnd1Out);   
  DP_Syn_RAM36x512    TpDrnd2
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, TPCirBuf2Out, TpDrnd2Out);
  DP_Syn_RAM36x512    TpDrnd3
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, TPCirBuf3Out, TpDrnd3Out);
  DP_Syn_RAM36x512    TpDrnd4
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, TPCirBuf4Out, TpDrnd4Out);
  DP_Syn_RAM36x512    TpDrnd5
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, TPCirBuf5Out, TpDrnd5Out);
  DP_Syn_RAM36x512    TpDrnd6
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, TPCirBuf6Out, TpDrnd6Out);
  DP_Syn_RAM36x512    TpDrnd7
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, TPCirBuf7Out, TpDrnd7Out);
  DP_Syn_RAM36x512    TpDrnd8
   (clk,DrndWe, DrndWrAddr, DrndRdAddr, TPCirBuf8Out, TpDrnd8Out);
   
// MULTIPLEXER
//  ID scheme:  channel id[4:0] : {fiber_id[4:2],channel_per_fiber[1:0]} 
 always @(posedge clk)   begin   
    case(ChanSel[5:0])

     6'd0:    DaqMuxOut <= {5'b00000, TpDrnd1Out[10:0]};
     6'd1:    DaqMuxOut <= {5'b00001, TpDrnd1Out[21:11]};
     6'd2:    DaqMuxOut <= {5'b00010, TpDrnd1Out[32:22]};
     6'd3:    DaqMuxOut <= {5'b00100, TpDrnd2Out[10:0]};
     6'd4:    DaqMuxOut <= {5'b00101, TpDrnd2Out[21:11]};
     6'd5:    DaqMuxOut <= {5'b00110, TpDrnd2Out[32:22]};
     6'd6:    DaqMuxOut <= {5'b01000, TpDrnd3Out[10:0]};
     6'd7:    DaqMuxOut <= {5'b01001, TpDrnd3Out[21:11]};  
     6'd8:    DaqMuxOut <= {5'b01010, TpDrnd3Out[32:22]};
     6'd9:    DaqMuxOut <= {5'b01100, TpDrnd4Out[10:0]};
     6'd10:   DaqMuxOut <= {5'b01101, TpDrnd4Out[21:11]};
     6'd11:   DaqMuxOut <= {5'b01110, TpDrnd4Out[32:22]};
     6'd12:   DaqMuxOut <= {5'b10000, TpDrnd5Out[10:0]};
     6'd13:   DaqMuxOut <= {5'b10001, TpDrnd5Out[21:11]};
     6'd14:   DaqMuxOut <= {5'b10010, TpDrnd5Out[32:22]};
     6'd15:   DaqMuxOut <= {5'b10100, TpDrnd6Out[10:0]};
     6'd16:   DaqMuxOut <= {5'b10101, TpDrnd6Out[21:11]};
     6'd17:   DaqMuxOut <= {5'b10110, TpDrnd6Out[32:22]};
     6'd18:   DaqMuxOut <= {5'b11000, TpDrnd7Out[10:0]};
     6'd19:   DaqMuxOut <= {5'b11001, TpDrnd7Out[21:11]};
     6'd20:   DaqMuxOut <= {5'b11010, TpDrnd7Out[32:22]};
     6'd21:   DaqMuxOut <= {5'b11100, TpDrnd8Out[10:0]};
     6'd22:   DaqMuxOut <= {5'b11101, TpDrnd8Out[21:11]};
     6'd23:   DaqMuxOut <= {5'b11110, TpDrnd8Out[32:22]};   
      
     6'd24:   DaqMuxOut <= {5'b00000, DaqDrnd1Out[10:0]};
     6'd25:   DaqMuxOut <= {5'b00001, DaqDrnd1Out[21:11]};
     6'd26:   DaqMuxOut <= {5'b00010, DaqDrnd1Out[32:22]};
     6'd27:   DaqMuxOut <= {5'b00100, DaqDrnd2Out[10:0]};
     6'd28:   DaqMuxOut <= {5'b00101, DaqDrnd2Out[21:11]};
     6'd29:   DaqMuxOut <= {5'b00110, DaqDrnd2Out[32:22]};
     6'd30:   DaqMuxOut <= {5'b01000, DaqDrnd3Out[10:0]};
     6'd31:   DaqMuxOut <= {5'b01001, DaqDrnd3Out[21:11]};  
     6'd32:   DaqMuxOut <= {5'b01010, DaqDrnd3Out[32:22]};
     6'd33:   DaqMuxOut <= {5'b01100, DaqDrnd4Out[10:0]};
     6'd34:   DaqMuxOut <= {5'b01101, DaqDrnd4Out[21:11]};
     6'd35:   DaqMuxOut <= {5'b01110, DaqDrnd4Out[32:22]};
     6'd36:   DaqMuxOut <= {5'b10000, DaqDrnd5Out[10:0]};
     6'd37:   DaqMuxOut <= {5'b10001, DaqDrnd5Out[21:11]};
     6'd38:   DaqMuxOut <= {5'b10010, DaqDrnd5Out[32:22]};
     6'd39:   DaqMuxOut <= {5'b10100, DaqDrnd6Out[10:0]};
     6'd40:   DaqMuxOut <= {5'b10101, DaqDrnd6Out[21:11]};
     6'd41:   DaqMuxOut <= {5'b10110, DaqDrnd6Out[32:22]};
     6'd42:   DaqMuxOut <= {5'b11000, DaqDrnd7Out[10:0]};
     6'd43:   DaqMuxOut <= {5'b11001, DaqDrnd7Out[21:11]};
     6'd44:   DaqMuxOut <= {5'b11010, DaqDrnd7Out[32:22]};
     6'd45:   DaqMuxOut <= {5'b11100, DaqDrnd8Out[10:0]};
     6'd46:   DaqMuxOut <= {5'b11101, DaqDrnd8Out[21:11]};
     6'd47:   DaqMuxOut <= {5'b11110, DaqDrnd8Out[32:22]};
     default    DaqMuxOut <=  16'd0;
    endcase
 end     // always

 always@(posedge clk)    begin     
  // + 2 ensures a proper behavior for any (PipeLength + NPS) >=0
        CircBufWrAd <= CircBufRdAd + PipeLength + NPS + 2;                
        CircBufRdAd <= CircBufRdAd + 1;
        DInErDel <= TpDrnd1Out[33];
 end       

endmodule