`timescale 1ns / 1ps

module sfifo36x512
     (   clk,      rst,
   		push,     pop,
	   	data_in,  data_out,
		   full,     empty    );

parameter   // do not change from higher levels
  DATA_WIDTH =  36,
  ADDR_WIDTH =   9,
  FIFO_DEPTH = 512;

input	  clk, 	rst;	
input	  push, pop;	 
input   [DATA_WIDTH-1:0]	data_in; 	
output  full,	empty;	
output  [DATA_WIDTH-1:0]	data_out;	

wire	clk,  rst;
wire	pop,   push;
wire  [DATA_WIDTH-1:0]	data_in;
wire  full,  empty;
						
reg [ADDR_WIDTH-1:0]	counter;  // # of occupied locations

reg [ADDR_WIDTH-1:0]	rd_pointer, wr_pointer;
wire logic0,logic1;
wire [DATA_WIDTH-1:0] logic_bus_0;
    
assign logic0 = 0;
assign logic1 = 1;
assign logic_bus_0 = 16'b0;
assign  full = (counter == FIFO_DEPTH-1) ? 1'b1 : 1'b0;
assign  empty = (counter == 0) ? 1'b1 : 1'b0;   // empty could be a reg 

	
RAMB16_S36_S36 bram1 ( .ADDRA(wr_pointer),		 .ADDRB(rd_pointer), 
                       .DIA(data_in[35:4]), .DIPA(data_in[3:0]), 
                       .DIB(logic_bus_0[35:4]), .DIPB(logic_bus_0[3:0]), 
                       .WEA(push), .WEB(logic0), .CLKA(clk), .CLKB(clk), 
                       .SSRA(logic0), .SSRB(logic0), .ENA(logic1), 
                       .ENB(logic1), .DOA(),
                       .DOPA(), .DOB(data_out[35:4]), .DOPB(data_out[3:0]) );

always @(posedge clk or posedge rst ) begin   //  always
	if (rst) begin		// Reset 
		rd_pointer <= 'b0;
		wr_pointer <= 'b0;
		counter    <= 'b0;
	end	           // Reset 
	else begin   // clk
		if (pop) begin    //  pop
			// Check for FIFO underflow
			if (counter == 0) begin  // empty case
			//	$display("\nERROR at time %0t:", $time);
			//	$display("FIFO Underflow\n");
				counter <= counter;			
				// Use $stop for debugging
			//	$stop;
			end     // empty case
			else begin     // non-empty case
  			 // no change to the counter if simultaneous read and write

			 // Decrement the FIFO counter if only pop
			 if (~push) 	counter <= counter - 1;

			 // Increment circularly the read pointer
			 if (rd_pointer == FIFO_DEPTH-1)    rd_pointer <= 'b0;
			 else                    rd_pointer <= rd_pointer + 1;
	    	end           // non-empty case		
		end        //  pop
		if (push) begin      //   push
			if (counter >= FIFO_DEPTH -1) begin	  //  overflow case
			$display("\nsfifo16x255 OVERFLOW at time %0t:", $time);
				counter <= counter;
				rd_pointer <=  rd_pointer + 1;
				wr_pointer <=  wr_pointer + 1;  // overwrite here
			
			$stop;	// Use $stop for debugging
			end         //  overflow case
			else begin       //  non-overflow case
			// If simultaneous read and write, no change to the counter

			// Increment the FIFO counter if only push
			if (~pop)	counter <= counter + 1;

			// Increment circularly the write pointer
			if (wr_pointer == FIFO_DEPTH-1)	wr_pointer <= 'b0;
			else                 wr_pointer <=  wr_pointer + 1;
			end            //  non-overflow case	
		end             //   push
	end        // clk
end      //  always
endmodule